from base_aux.base_lambdas.m1_lambda import *
from base_aux.base_nest_dunders.m7_cmp import *
import operator

# =====================================================================================================================
class Victim(NestCmp_GLET_DigitAccuracy):
    def __init__(self, source: float | int, **kwargs):
        self.SOURCE = source
        super().__init__(**kwargs)

    @property
    def CMP_VALUE(self) -> TYPING.DIGIT_FLOAT_INT:
        return self.SOURCE


# =====================================================================================================================
class Test_CmpDigit:
    @pytest.mark.parametrize(
        argnames="source, accuracy_value, other, _EXPECTED",
        argvalues=[
            (1, None, 1, (True, False, True, False)),
            (1, None, "hello", (Exception, Exception, Exception, Exception)),
        ]
    )
    def test__cmp_glet(
            self,
            source: float | int,
            accuracy: float | None,
            other: float | int,
            _EXPECTED: bool | Exception,
    ):
        # -----------------------------------------------
        victim = Victim(source=source, cmp_accuracy_value=accuracy)

        Lambda(victim.cmp_ge, other=other).check_expected__assert(_EXPECTED[0])
        Lambda(victim.cmp_gt, other=other).check_expected__assert(_EXPECTED[1])
        Lambda(victim.cmp_le, other=other).check_expected__assert(_EXPECTED[2])
        Lambda(victim.cmp_lt, other=other).check_expected__assert(_EXPECTED[3])

        # -----------------------------------------------
        victim = Victim(source=source)

        Lambda(victim.cmp_ge, other=other, accuracy=accuracy).check_expected__assert(_EXPECTED[0])
        Lambda(victim.cmp_gt, other=other, accuracy=accuracy).check_expected__assert(_EXPECTED[1])
        Lambda(victim.cmp_le, other=other, accuracy=accuracy).check_expected__assert(_EXPECTED[2])
        Lambda(victim.cmp_lt, other=other, accuracy=accuracy).check_expected__assert(_EXPECTED[3])

        # -----------------------------------------------
        victim = Victim(source=source, cmp_accuracy_value=accuracy)

        Lambda(victim.cmp_ge, other=other, accuracy=accuracy).check_expected__assert(_EXPECTED[0])
        Lambda(victim.cmp_gt, other=other, accuracy=accuracy).check_expected__assert(_EXPECTED[1])
        Lambda(victim.cmp_le, other=other, accuracy=accuracy).check_expected__assert(_EXPECTED[2])
        Lambda(victim.cmp_lt, other=other, accuracy=accuracy).check_expected__assert(_EXPECTED[3])

        # -----------------------------------------------
        # try:      # NOTE: USE OPERATOR instead!!!
        #     result = victim >= other
        # except Exception as exc:
        #     TypeAux(exc).check__subclassed_or_isinst__from_cls_or_inst(_EXPECTED[0])
        # else:
        #     assert result is _EXPECTED[0]

        victim = Victim(source=source, cmp_accuracy_value=accuracy)

        Lambda(operator.ge, victim, other).check_expected__assert(_EXPECTED[0])
        Lambda(operator.gt, victim, other).check_expected__assert(_EXPECTED[1])
        Lambda(operator.le, victim, other).check_expected__assert(_EXPECTED[2])
        Lambda(operator.lt, victim, other).check_expected__assert(_EXPECTED[3])


# =====================================================================================================================



import pytest

from base_aux.base_lambdas.m1_lambda import *
from base_aux.base_nest_dunders.m4_gsai_annots import *


# =====================================================================================================================
class Test__Attr:
    @classmethod
    def setup_class(cls):
        pass

        class Victim(NestGAI_AnnotAttrIC):
            attr_lowercase = "value"
            ATTR_UPPERCASE = "VALUE"
            Attr_CamelCase = "Value"
            # NOT_EXISTS

        cls.Victim = Victim

    @classmethod
    def teardown_class(cls):
        pass

    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.parametrize(
        argnames="attr, _EXPECTED",
        argvalues=[
            (0, Exception),
            (1, Exception),
            (None, Exception),
            (True, Exception),
            ("", Exception),
            (" TRUE", Exception),
            ("NOT_EXISTS", Exception),

            ("__name__", "Victim"),

            ("attr_lowercase", "value"),
            ("ATTR_LOWERCASE", "value"),

            ("ATTR_UPPERCASE", "VALUE"),
            ("attr_uppercase", "VALUE"),

            ("     attr_uppercase", "VALUE"),
        ]
    )
    def test__get(self, attr, _EXPECTED):
        Lambda(lambda: getattr(self.Victim(), attr)).check_expected__assert(_EXPECTED)
        Lambda(lambda: self.Victim()[attr]).check_expected__assert(_EXPECTED)

    # @pytest.mark.parametrize(
    #     argnames="attr, _EXPECTED",
    #     argvalues=[
    #         (None, Exception),
    #         (True, Exception),
    #         ("", Exception),
    #         (" TRUE", Exception),
    #
    #         ("attr_lowercase", None),
    #         ("ATTR_LOWERCASE", None),
    #
    #         ("ATTR_UPPERCASE", None),
    #         ("attr_uppercase", None),
    #
    #         ("     attr_uppercase", None),
    #     ]
    # )
    # def test__set(self, attr, _EXPECTED):
    #     NEW_VALUE = "NEW_VALUE"
    #     victim = Victim()
    #     Lambda(lambda: setattr(victim, attr, NEW_VALUE)).check_assert(_EXPECTED)
    #     if _EXPECTED == Exception:
    #         return
    #     Lambda(lambda: getattr(victim, attr)).check_assert(NEW_VALUE)
    #     # Lambda(lambda: Victim()[attr] = NEW_VALUE).check_assert(_EXPECTED)


# =====================================================================================================================
class Test__Annot:
    @classmethod
    def setup_class(cls):
        pass

        class Victim(NestGAI_AnnotAttrIC):
            attr_lowercase: str = "value"
            ATTR_UPPERCASE = "VALUE"
            Attr_CamelCase = "Value"
            # NOT_EXISTS

        cls.Victim = Victim

    @classmethod
    def teardown_class(cls):
        pass

    # -----------------------------------------------------------------------------------------------------------------
    @pytest.mark.parametrize(
        argnames="attr, _EXPECTED",
        argvalues=[
            (0, "value"),
            (1, Exception),
            (None, Exception),
            (True, Exception),
            ("", Exception),
            (" TRUE", Exception),
            ("NOT_EXISTS", Exception),

            ("__name__", "Victim"),

            ("attr_lowercase", "value"),
            ("ATTR_LOWERCASE", "value"),

            ("ATTR_UPPERCASE", "VALUE"),
            ("attr_uppercase", "VALUE"),

            ("     attr_uppercase", "VALUE"),
        ]
    )
    def test__get(self, attr, _EXPECTED):
        # Lambda(lambda: getattr(self.Victim(), attr)).check_assert(_EXPECTED)
        Lambda(lambda: self.Victim()[attr]).check_expected__assert(_EXPECTED)


# =====================================================================================================================
def test__cls_name():
    assert NestGAI_AnnotAttrIC().__class__.__name__ == f"NestGAI_AnnotAttrIC"

    class Victim(NestGAI_AnnotAttrIC):
        A1: int = 1

    assert Victim().__class__.__name__ == f"Victim"


# =====================================================================================================================



from typing import *
from PyQt5.QtCore import QThread

from base_aux.base_lambdas.m1_lambda import *


# =====================================================================================================================
class LambdaThread(Lambda, QThread):
    """
    NOTE
    ----
    same as Lambda but just add nesting QThread!

    GOAL
    ----
    Object for keeping thread data for better managing.
    """
    def __SLOTS_EXAMPLES(self):
        """DON'T START! just for explore!
        """
        # checkers --------------------
        self.started
        self.isRunning()

        self.finished
        self.isFinished()

        self.destroyed
        self.signalsBlocked()

        # settings -------------------
        self.setTerminationEnabled()

        # NESTING --------------------
        self.currentThread()
        self.currentThreadId()
        self.thread()
        self.children()
        self.parent()

        # info --------------------
        self.priority()
        self.loopLevel()
        self.stackSize()
        self.idealThreadCount()

        self.setPriority()
        self.setProperty()
        self.setObjectName()

        self.tr()

        self.dumpObjectInfo()
        self.dumpObjectTree()

        # CONTROL --------------------
        self.run()
        self.start()
        self.startTimer()

        self.sleep(100)
        self.msleep(100)
        self.usleep(100)

        self.wait()

        self.killTimer()

        self.disconnect()
        self.deleteLater()
        self.terminate()
        self.quit()
        self.exit(100)

        # WTF --------------------


# =====================================================================================================================



import pytest
from base_aux.bits.m1_bits import *


# =====================================================================================================================
class Test__Bitfield:
    VICTIM: type[Bitfield] = type("VICTIM", (Bitfield,), {})

    @classmethod
    def setup_class(cls):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    def setup_method(self, method):
        self.VICTIM = type("VICTIM", (Bitfield,), {})

    # -----------------------------------------------------------------------------------------------------------------
    def test__0_bits(self):
        try:
            self.VICTIM(0)
        except Exc__Incompatible_Data:
            assert True
        else:
            assert False

    def test__1_bits(self):
        victim: Bitfield = self.VICTIM(1)
        assert victim.field_size == 1
        assert len(victim) == 1
        assert victim._field_bytearray == bytearray(1)

        # [0] -------------------------------------------
        victim[0] = 0

        assert victim[0] == 0
        assert victim.int() == 0
        assert victim.count_activated() == 0

        assert victim.field_str == "0"
        assert victim._bin_str() == "0"
        assert victim._hex_str() == "00"
        assert str(victim) == "field[0]"
        assert victim.size_get_active() == 0

        assert victim.list_bits() == [0, ]
        assert victim.list_activated_indexes() == []
        assert victim.list_activated_values() == []

        # [1] -------------------------------------------
        victim[0] = 1

        assert victim[0] == 1
        assert victim.int() == 1
        assert victim.count_activated() == 1

        assert victim.field_str == "1"
        assert victim._bin_str() == "1"
        assert victim._hex_str() == "01"
        assert str(victim) == "field[1]"
        assert victim.size_get_active() == 1

        assert victim.list_bits() == [1, ]
        assert victim.list_activated_indexes() == [0, ]
        assert victim.list_activated_values() == [1, ]

    def test__2_bits(self):
        victim: Bitfield = self.VICTIM(2)
        assert victim.field_size == 2
        assert len(victim) == 2
        assert victim._field_bytearray == bytearray(1)

        # [00] -------------------------------------------
        victim[0] = 0
        victim[1] = 0

        assert victim[0] == 0
        assert victim.int() == 0
        assert victim.count_activated() == 0

        assert victim.field_str == "00"
        assert victim._bin_str() == "00"
        assert victim._hex_str() == "00"
        assert str(victim) == "field[00]"
        assert victim.size_get_active() == 0

        assert victim.list_bits() == [0, 0, ]
        assert victim.list_activated_indexes() == []
        assert victim.list_activated_values() == []

        # [01] -------------------------------------------
        victim[0] = 1
        victim[1] = 0

        assert victim[0] == 1
        assert victim.int() == 1
        assert victim.count_activated() == 1

        assert victim.field_str == "01"
        assert victim._bin_str() == "01"
        assert victim._hex_str() == "01"
        assert str(victim) == "field[01]"
        assert victim.size_get_active() == 1

        assert victim.list_bits() == [0, 1, ]
        assert victim.list_activated_indexes() == [0, ]
        assert victim.list_activated_values() == [1, ]

        # [10] -------------------------------------------
        victim[0] = 0
        victim[1] = 1

        assert victim[0] == 0
        assert victim[1] == 1
        assert victim.int() == 2
        assert victim.count_activated() == 1

        assert victim.field_str == "10"
        assert victim._bin_str() == "10"
        assert victim._hex_str() == "02"
        assert str(victim) == "field[10]"
        assert victim.size_get_active() == 2

        assert victim.list_bits() == [1, 0, ]
        assert victim.list_activated_indexes() == [1, ]
        assert victim.list_activated_values() == [2, ]

    def test__index(self):
        pass


# =====================================================================================================================



from base_aux.valid.m1_valid_base import *
from base_aux.translator.m1_translator import Translator
from base_aux.pyqt.m3_tm import Base_TableModel
from base_aux.breeders.m1_breeder_str2_stack import *
from base_aux.breeders.m1_breeder_str1_series import *
from base_aux.pyqt.m0_static import *


# =====================================================================================================================
class TcResultMsg:
    PASS: str = "Успех"
    FAIL: str = "Ошибка"
    WAIT: str = "..."


# =====================================================================================================================
class TableModel_Tps(Base_TableModel):
    DATA: "TpManager"
    HEADERS: "Headers"
    HTRANSLATOR: Translator

    # AUX -------------------------------------------
    open__settings: Optional[bool] = None

    def __init__(self, data: "TpManager" = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.reinit(data)

    def reinit(self, data: "TpManager" = None) -> None:
        """
        GOAL
        ----
        separate inition!
        """
        if data is not None:
            self.DATA = data

        class Headers(BreederStrStack):
            TESTCASE: int = 0
            SKIP: None = None
            ASYNC: None = None
            STARTUP_CLS: None = None
            DUTS: BreederStrSeries = BreederStrSeries(None, self.DATA.STAND.DEV_LINES.COUNT_COLUMNS)
            TEARDOWN_CLS: None = None
            # FIXME: need resolve COUNT over DevicesIndexed!!!

        class HTRus:
            TESTCASE: str = "ТЕСТКЕЙС"
            SKIP: str = "Пропустить"
            ASYNC: str = "Асинхр."
            STARTUP_CLS: str = "Подготовка\nтесткейса"
            TEARDOWN_CLS: str = "Завершение\nтесткейса"

        self.HEADERS = Headers()
        self.HTRANSLATOR = Translator(HTRus)

    def rowCount(self, parent: QModelIndex = None, *args, **kwargs) -> int:
        return len(self.DATA.STAND.TCSc_LINE) + 1  # [+1]for finalResults

    def columnCount(self, parent: QModelIndex = None, *args, **kwargs) -> int:
        return self.HEADERS.count()

    # def headerData(self, section: Any, orientation: Qt.Orientation, role: int = Qt.DisplayRole) -> str:
    #     if role == Qt.DisplayRole:
    #         # ------------------------------
    #         if orientation == Qt.Horizontal:
    #             return self.HEADERS[section]
    #
    #         # ------------------------------
    #         if orientation == Qt.Vertical:
    #             return str(section + 1)

    def flags(self, index: QModelIndex) -> int:
        # PREPARE -----------------------------------------------------------------------------------------------------
        col = index.column()
        row = index.row()

        try:
            tc_cls = list(self.DATA.STAND.TCSc_LINE)[row]
        except:
            tc_cls = None

        row_is_summary: bool = tc_cls is None

        # -------------------------------------------------------------------------------------------------------------
        flags = super().flags(index)

        if row_is_summary:
            pass

        elif col in [self.HEADERS.SKIP, self.HEADERS.ASYNC] or col in self.HEADERS.DUTS:
            flags |= Qt.ItemIsUserCheckable
            # flags |= Qt.ItemIsSelectable
        else:
            # flags -= Qt.ItemIsSelectable
            pass

        # clear SELECTABLE ---------
        return flags

    def get_summary_results__str(self, col: int) -> list[str]:
        result = []
        for row in range(self.rowCount() - 1):
            result_i = self.index(row, col).data()
            result.append(result_i)

        # print(f"{col=}/{result=}")
        return result

    def get_summary_result(self, col: int) -> bool | None:
        for row in range(self.rowCount() - 1):
            result_i = self.index(row, col).data()
            # print(f"{col=}/{row=}/{result_i=}")
            if result_i == TcResultMsg.PASS:
                continue
            if result_i == TcResultMsg.FAIL:
                return False
            if result_i == TcResultMsg.WAIT:
                return
            else:
                return
        return True

    def get_summary_result__str(self, col: int) -> str:
        result = self.get_summary_result(col)
        if result is True:
            return TcResultMsg.PASS
        elif result is False:
            return TcResultMsg.FAIL
        else:
            return ""

    def data(self, index: QModelIndex, role: int = Qt.DisplayRole) -> Any:
        # PREPARE -----------------------------------------------------------------------------------------------------
        col = index.column()
        row = index.row()

        try:
            tc_cls = list(self.DATA.STAND.TCSc_LINE)[row]
        except:
            tc_cls = None

        row_is_summary: bool = tc_cls is None

        dut = None
        tc_inst = None
        if col in self.HEADERS.DUTS and not row_is_summary:
            index = col - self.HEADERS.DUTS.START_OUTER
            try:
                dut = self.DATA.STAND.DEV_LINES.DUT[index]
                # print(f"{tc_cls.TCSi_LINE=}/{index=}")
                tc_inst = tc_cls.TCSi_LINE[index]
            except:
                return

        # -------------------------------------------------------------------------------------------------------------
        if role == Qt.DisplayRole:
            if col == self.HEADERS.TESTCASE:
                # return f"{tc_cls.NAME}\n{tc_cls.DESCRIPTION}"
                if row_is_summary:
                    return "ИТОГ:"
                return f"{tc_cls.DESCRIPTION}"
            if col == self.HEADERS.SKIP:
                if row_is_summary:
                    return
                return '+' if tc_cls.SKIP else '-'
            if col == self.HEADERS.ASYNC:
                if row_is_summary:
                    return
                return '+' if tc_cls.ASYNC else '-'

            # STARTUP -------------------
            if col == self.HEADERS.STARTUP_CLS:
                if row_is_summary:
                    return
                try:
                    group_name = tc_cls._EQ_CLS__VALUE.name or ""
                except:
                    group_name = tc_cls._EQ_CLS__VALUE or ""

                if tc_cls.result__startup_cls is None:
                    return group_name
                if tc_cls.result__startup_cls is EnumAdj_ProcessStateActive.STARTED:
                    return TcResultMsg.WAIT
                if isinstance(tc_cls.result__startup_cls, Valid) and tc_cls.result__startup_cls.check__active():
                    return TcResultMsg.WAIT
                if bool(tc_cls.result__startup_cls) is True:
                    return f'+{group_name}'
                if bool(tc_cls.result__startup_cls) is False:
                    return f'-{group_name}'

            # DUTS -------------------
            if col in self.HEADERS.DUTS:
                if row_is_summary:
                    return self.get_summary_result__str(col)
                if tc_inst:
                    if tc_inst.result is None:
                        return ""
                    elif isinstance(tc_inst.result, Valid):
                        if tc_inst.result.check__active():
                            return TcResultMsg.WAIT

                        elif tc_inst.result.STATE_ACTIVE == EnumAdj_ProcessStateActive.FINISHED:
                            if bool(tc_inst.result):
                                return TcResultMsg.PASS
                            else:
                                return TcResultMsg.FAIL
                        else:
                            return

                    else:
                        if bool(tc_inst.result):
                            return TcResultMsg.PASS
                        else:
                            return TcResultMsg.FAIL

            # TEARDOWN -------------------
            if col == self.HEADERS.TEARDOWN_CLS:
                if row_is_summary:
                    return
                if tc_cls.result__teardown_cls is None:
                    return
                if tc_cls.result__teardown_cls is EnumAdj_ProcessStateActive.STARTED:
                    return TcResultMsg.WAIT
                if isinstance(tc_cls.result__teardown_cls, Valid) and tc_cls.result__teardown_cls.check__active():
                    return TcResultMsg.WAIT
                if bool(tc_cls.result__teardown_cls) is True:
                    return TcResultMsg.PASS
                if bool(tc_cls.result__teardown_cls) is False:
                    return TcResultMsg.FAIL

        # -------------------------------------------------------------------------------------------------------------
        if role == Qt.ToolTipRole:
            if col == self.HEADERS.TESTCASE:
                if row_is_summary:
                    return "Результаты суммарные по всем тесткейсам\nдля каждого устройства"
                return f"{tc_cls.DESCRIPTION}"
            if col in self.HEADERS.DUTS:
                if row_is_summary:
                    # return
                    return str(self.get_summary_results__str(col))
                if tc_inst:
                    return f"{tc_inst.result}"

        # -------------------------------------------------------------------------------------------------------------
        if role == Qt.TextAlignmentRole:
            """
            VARIANTS ALIGN
            --------------
            not exists NAME!!!} = 0         # (LEFT+TOP) [[[[[[[[DEFAULT IS [LEFT+TOP]]]]]]]]]
            
            AlignLeft=AlignLeading = 1      # LEFT(+TOP)
            AlignRight=AlignTrailing = 2    # RIGHT(+TOP)

            AlignTop = 32       # TOP(+LEFT)
            AlignBottom = 64    # BOT(+LEFT)

            AlignHCenter = 4    # HCENTER(+TOP)
            AlignVCenter = 128  # VCENTER(+LEFT)
            AlignCenter = 132   # VCENTER+HCENTER

            # =====MAYBE DID NOT FIGURED OUT!!!
            AlignAbsolute = 16      # (LEFT+TOP) == asDEFAULT
            AlignBaseline = 256     # (LEFT+TOP) == asDEFAULT

            AlignJustify = 8        # (LEFT+TOP) == asDEFAULT

            AlignHorizontal_Mask = 31   # TOP+RIGHT
            AlignVertical_Mask = 480    # LEFT+VCENTER
            """
            if col == self.HEADERS.TESTCASE:
                if row_is_summary:
                    return ALIGNMENT.CR
                return ALIGNMENT.CL
            else:
                return ALIGNMENT.C

        # -------------------------------------------------------------------------------------------------------------
        if role == Qt.TextColorRole:
            if row_is_summary:
                return
            if tc_cls.SKIP:
                return QColor('#a2a2a2')

        # -------------------------------------------------------------------------------------------------------------
        if role == Qt.BackgroundColorRole:
            if tc_cls and tc_cls.SKIP:
                return QColor('#e2e2e2')

            # ACTIVE ---------------------
            if col == self.HEADERS.TESTCASE:
                if row_is_summary:
                    return
                if self.DATA.tc_active == tc_cls:
                    return QColor("#FFFF50")

            # STARTUP -------------------
            if col == self.HEADERS.STARTUP_CLS:
                if row_is_summary:
                    return
                if tc_cls.result__startup_cls is None:
                    return
                if tc_cls.result__startup_cls is EnumAdj_ProcessStateActive.STARTED:
                    return QColor("#FFFF50")
                if isinstance(tc_cls.result__startup_cls, Valid) and tc_cls.result__startup_cls.check__active():
                    return QColor("#FFFF50")
                if bool(tc_cls.result__startup_cls) is True:
                    return QColor("#50FF50")
                if bool(tc_cls.result__startup_cls) is False:
                    return QColor("#FF5050")

            # DUTS -------------------
            if col in self.HEADERS.DUTS:
                if row_is_summary:
                    result_i = self.get_summary_result(col)
                    if result_i is True:
                        return QColor("#00FF00")
                    elif result_i is False:
                        return QColor("#FF5050")
                    else:
                        return
                if tc_inst.skip_tc_dut or dut.SKIP or not dut.DEV_FOUND:
                    return QColor('#e2e2e2')
                if tc_inst.result__startup is not None and not bool(tc_inst.result__startup):
                    return QColor("#FFa0a0")
                if tc_inst.isRunning():
                    return QColor("#FFFF50")
                if bool(tc_inst.result) is True:
                    return QColor("#00FF00")
                if bool(tc_inst.result) is False:
                    if tc_inst.result is None:
                        return
                    else:
                        return QColor("#FF5050")
                # elif

            # TEARDOWN -------------------
            if col == self.HEADERS.TEARDOWN_CLS:
                if row_is_summary:
                    return
                if tc_cls.result__teardown_cls is None:
                    return
                if tc_cls.result__teardown_cls is EnumAdj_ProcessStateActive.STARTED:
                    return QColor("#FFFF50")
                if isinstance(tc_cls.result__teardown_cls, Valid) and tc_cls.result__teardown_cls.check__active():
                    return QColor("#FFFF50")
                if bool(tc_cls.result__teardown_cls) is True:
                    return QColor("#50FF50")
                if bool(tc_cls.result__teardown_cls) is False:
                    return QColor("#FF5050")

        # -------------------------------------------------------------------------------------------------------------
        if role == Qt.CheckStateRole:
            # -------------------------
            if row_is_summary:
                return
            if not self.open__settings:
                return

            # -------------------------
            if col == self.HEADERS.SKIP:
                if tc_cls.SKIP:
                    return Qt.Checked
                else:
                    return Qt.Unchecked

            if col == self.HEADERS.ASYNC:
                if tc_cls.ASYNC:
                    return Qt.Checked
                else:
                    return Qt.Unchecked

            if col in self.HEADERS.DUTS:
                if (tc_cls and tc_cls.SKIP) or (dut and dut.SKIP):
                    return

                if tc_inst.skip_tc_dut:
                    return Qt.Unchecked
                else:
                    return Qt.Checked

        # -------------------------------------------------------------------------------------------------------------
        if role == Qt.FontRole:
            if tc_cls == self.DATA.tc_active:
                # QFont("Arial", 9, QFont.Bold)
                font = QFont()

                font.setBold(True)
                # font.setItalic(True)

                # font.setOverline(True)  # надчеркнутый
                # font.setStrikeOut(True)  # зачеркнутый
                # font.setUnderline(True)  # подчеркнутый

                # не понял!! --------------------
                # font.setStretch(5)
                # font.setCapitalization()

                return font

    def setData(self, index: QModelIndex, value: Any, role: int = None) -> bool:
        # PREPARE -----------------------------------------------------------------------------------------------------
        col = index.column()
        row = index.row()

        try:
            tc_cls = list(self.DATA.STAND.TCSc_LINE)[row]
        except:
            tc_cls = None

        row_is_summary: bool = tc_cls is None

        dut = None
        tc_inst = None
        if col in self.HEADERS.DUTS and not row_is_summary:
            index = col - self.HEADERS.DUTS.START_OUTER
            try:
                dut = self.DATA.STAND.DEV_LINES.DUT[index]
                # print(f"{tc_cls.TCSi_LINE=}/{index=}")
                tc_inst = tc_cls.TCSi_LINE[index]
            except:
                return True

        # -------------------------------------------------------------------------------------------------------------
        if role == Qt.CheckStateRole:
            if row_is_summary:
                return True
            if col == self.HEADERS.SKIP:
                tc_cls.SKIP = value == Qt.Checked

            if col == self.HEADERS.ASYNC:
                tc_cls.ASYNC = value == Qt.Checked

            if col in self.HEADERS.DUTS:
                if tc_inst:
                    tc_inst.skip_tc_dut = value == Qt.Unchecked

        # FINAL -------------------------------------------------------------------------------------------------------
        self._data_reread()
        return True


# =====================================================================================================================



import sys
from typing import *

from base_aux.base_nest_dunders.m3_bool import NestBool_False


# =====================================================================================================================
class Print:
    """
    GOAL
    ----
    print multy lines msg in stdout

    SPECIALLY CREATED FOR
    ---------------------
    Base for Warn

    TODO: try use direct logger?
        or rename into some new class! as universal Msging!
    """
    OUTPUT: TextIO = sys.stdout
    PREFIX: str = "[INFO]"

    INDENT: str = "__"
    EOL: str = "\n"
    MSG_LINES: tuple[str, ...]

    def __init__(self, *lines, prefix: str = None, **kwargs) -> None:
        if prefix is not None:
            self.PREFIX = prefix

        self.MSG_LINES = lines
        print(self, file=self.OUTPUT)

        if self.OUTPUT == sys.stderr:
            # additional print - cause of stderr prints not in same
            print(self, file=sys.stdout)

        super().__init__(**kwargs)

    def __str__(self):
        return self.MSG_STR

    @property
    def MSG_STR(self) -> str:
        result = f"{self.PREFIX}"
        for index, line in enumerate(self.MSG_LINES):
            if index == 0:
                result += f"{line}"
            else:
                result += f"{self.EOL}{self.INDENT}{line}"

        return result


# =====================================================================================================================
class Warn(
    Print,

    NestBool_False,
):
    """
    GOAL
    ----
    when you dont want to use logger and raise error (by now).
    print msg in some inner functions when raising Exc after inner function return False.

    SPECIALLY CREATED FOR
    ---------------------
    ReleaseHistory.check_new_release__is_correct/generate
    """
    OUTPUT: TextIO = sys.stderr
    PREFIX: str = "[WARN]"


# =====================================================================================================================



from typing import *

from base_aux.base_nest_dunders.m1_init1_source import *
from base_aux.aux_attr.m4_kits import *
from base_aux.base_nest_dunders.m3_calls import *
from base_aux.loggers.m1_print import *


# =====================================================================================================================
class ChainResolve(NestInit_Source, NestCall_Resolve, Base_AttrKit):
    """
    GOAL
    ----
    get cumulated result from several resolvers/callers

    SPECIALLY CREATED FOR
    ---------------------
    HtmlTag to find text by chain
    """
    SOURCE: Any = None
    CHAINS: tuple[Callable | NestCall_Resolve, ...] = ()

    ON_RAISE: EnumAdj_ReturnOnRaise = EnumAdj_ReturnOnRaise.NONE

    def __init__(self, *chains, source: Any = None, **kwargs):
        if chains:
            self.CHAINS = chains
        super().__init__(source=source, **kwargs)

    def resolve(self, source: Any = None, **kwargs) -> Any | NoReturn:
        if source in [None, NoValue]:
            source = self.SOURCE

        for chain in self.CHAINS:
            try:
                source = chain(source)
            except Exception as exc:
                Warn(f"{source=}/{exc!r}")

                if self.ON_RAISE == EnumAdj_ReturnOnRaise.NONE:
                    return None
                elif self.ON_RAISE == EnumAdj_ReturnOnRaise.RAISE:
                    raise exc
                elif self.ON_RAISE == EnumAdj_ReturnOnRaise.EXC:    # NOTE: dont use it! just as variant
                    return exc

        return source


# =====================================================================================================================



import pytest

from base_aux.base_lambdas.m1_lambda import *
from base_aux.base_values.m4_primitives import *
from base_aux.aux_eq.m3_eq_valid3_derivatives import *


# =====================================================================================================================
def test__reinit__mutable_cls_values():
    class Victim:
        LIST = []
        SET = {}
        DICT = {}

        def __init__(self, LIST=None, SET=None, DICT=None):
            if LIST is not None:
                self.LIST = LIST
            if SET is not None:
                self.SET = SET
            if DICT is not None:
                self.DICT = DICT

    # SAME BLANK ----------
    victim = Victim()

    assert Victim.LIST == victim.LIST
    assert Victim.SET == victim.SET
    assert Victim.DICT == victim.DICT

    assert Victim.LIST is victim.LIST
    assert Victim.SET is victim.SET
    assert Victim.DICT is victim.DICT

    AttrAux_Existed(victim).reinit__mutable_cls_values()
    assert Victim.LIST is not victim.LIST
    assert Victim.SET is not victim.SET
    assert Victim.DICT is not victim.DICT

    # DIFF ----------
    victim = Victim([], {}, {})
    assert Victim.LIST == victim.LIST
    assert Victim.SET == victim.SET
    assert Victim.DICT == victim.DICT

    assert Victim.LIST is not victim.LIST
    assert Victim.SET is not victim.SET
    assert Victim.DICT is not victim.DICT

    AttrAux_Existed(victim).reinit__mutable_cls_values()
    assert Victim.LIST is not victim.LIST
    assert Victim.SET is not victim.SET
    assert Victim.DICT is not victim.DICT


# =====================================================================================================================
class Victim:
    a=1
    _h=2
    __p=3
    __name__=123
    __hello__=123


class VictimNested_Old(Victim):
    pass


class VictimNested_ReNew(Victim):
    a = 1
    _h = 2
    __p = 3
    __name__=123
    __hello__=123


class VictimNested_New(Victim):
    a2 = 1
    _h2 = 2
    __p2 = 3
    __name__=123
    __hello__=123


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="source, _EXPECTED",
    argvalues=[
        (Victim,    ({"a", "_h", "__p"}, {"a", }, {"a", "_h", }, {"__p", })),
        (Victim(),  ({"a", "_h", "__p"}, {"a", }, {"a", "_h", }, {"__p", })),

        (VictimNested_Old,      ({"a", "_h", "__p"}, {"a", }, {"a", "_h", }, {"__p", })),
        (VictimNested_Old(),    ({"a", "_h", "__p"}, {"a", }, {"a", "_h", }, {"__p", })),

        (VictimNested_ReNew,    ({"a", "_h", "__p"}, {"a", }, {"a", "_h", }, {"__p", })),
        (VictimNested_ReNew(),  ({"a", "_h", "__p"}, {"a", }, {"a", "_h", }, {"__p", })),

        (VictimNested_New,
         (
                {"a", "_h", "__p", "a2", "_h2", "__p2", },
                {"a", "a2", },
                {"a", "_h", "a2", "_h2", },
                {"__p", "__p2", },
         )),
        (VictimNested_New(),
         (
                 {"a", "_h", "__p", "a2", "_h2", "__p2", },
                 {"a", "a2", },
                 {"a", "_h", "a2", "_h2", },
                 {"__p", "__p2", },
         )),
    ]
)
def test__iter(source, _EXPECTED):
    Lambda(set(AttrAux_Existed(source).iter__dirnames_original_not_builtin())).check_expected__assert(_EXPECTED[0])
    Lambda(set(AttrAux_Existed(source).iter__names_filter__not_hidden())).check_expected__assert(_EXPECTED[1])
    Lambda(set(AttrAux_Existed(source).iter__names_filter__not_private())).check_expected__assert(_EXPECTED[2])
    Lambda(set(AttrAux_Existed(source).iter__names_filter__private())).check_expected__assert(_EXPECTED[3])


# =====================================================================================================================
class Victim2:
    attr_lowercase = "value"
    ATTR_UPPERCASE = "VALUE"
    Attr_CamelCase = "Value"


class Victim3:
    AnE: int = 1
    AnNE: int


@pytest.mark.parametrize(
    argnames="victim, attr, _EXPECTED",
    argvalues=[
        (Victim2(), 1, (Exception, Exception, Exception, Exception, )),
        (Victim2(), None, (None, False, Exception, Exception, )),
        (Victim2(), True, (None, False, Exception, Exception, )),
        (Victim2(), "", (None, False, Exception, Exception, )),
        (Victim2(), " TRUE", (None, False, Exception, None, )),

        (Victim2(), "attr_lowercase", ("attr_lowercase", True, "value", None)),
        (Victim2(), "ATTR_LOWERCASE", ("attr_lowercase", True, "value", None, )),

        (Victim2(), "ATTR_UPPERCASE", ("ATTR_UPPERCASE", True, "VALUE", None, )),
        (Victim2(), "attr_uppercase", ("ATTR_UPPERCASE", True, "VALUE", None, )),

        (Victim2(), "     attr_uppercase", ("ATTR_UPPERCASE", True, "VALUE", None, )),

        # ANNOTS
        (Victim3(), "AnE", ("AnE", True, 1, None, )),
        (Victim3(), "ANE", ("AnE", True, 1, None, )),
        # (Victim3(), "AnNE", ("AnNE", True, Exception, None, )),   # FIXME: !!!!!!!!
        # (Victim3(), "ANNE", ("AnNE", True, Exception, None, )),   # FIXME: !!!!!!!!
    ]
)
def test__gsai(victim, attr, _EXPECTED):
    # use here EXACTLY the instance! if used class - value would changed in class and further values will not cmp correctly!

    Lambda(AttrAux_Existed(victim).name_ic__get_original, attr).check_expected__assert(_EXPECTED[0])
    Lambda(AttrAux_Existed(victim).name_ic__check_exists, attr).check_expected__assert(_EXPECTED[1])
    Lambda(AttrAux_Existed(victim).gai_ic, attr).check_expected__assert(_EXPECTED[2])
    Lambda(AttrAux_Existed(victim).sai_ic, attr, 123).check_expected__assert(_EXPECTED[3])


# =====================================================================================================================
def test__kwargs():
    victim = AttrDumped()
    AttrAux_Existed(victim).sai__by_args_kwargs(**dict(a1=1, A2=2))
    assert victim.a1 == 1
    assert victim.A2 == 2

    class Victim:
        a2=2

    victim = Victim()
    assert not hasattr(victim, "a1")
    assert hasattr(victim, "a2")
    assert not hasattr(victim, "A2")

    AttrAux_Existed(victim).sai__by_args_kwargs(**dict(a1=1))
    assert victim.a1 == 1

    AttrAux_Existed(victim).sai__by_args_kwargs(**dict(A2=22))
    assert victim.a2 == 22

    assert not hasattr(victim, "A2")


# =====================================================================================================================
class Victim:
    NONE = None
    TRUE = True
    LTRUE = LAMBDA_TRUE
    RAISE = LAMBDA_RAISE


class VictimNames:
    attr = None
    _attr = None
    __attr = None


class Test__Dump:
    def test__zero(self):
        assert AttrAux_Existed().dump_dict() == dict()

    @pytest.mark.parametrize(
        argnames="source, skip, _EXPECTED",
        argvalues=[
            (VictimNames(), [], {"attr": None, "_attr": None}),
            (VictimNames(), ["attr", ], {"_attr": None}),
            (VictimNames(), [EqValid_Contain("att"), ], {}),
            (VictimNames(), [EqValid_Contain("att5"), ], {"attr": None, "_attr": None}),
            (VictimNames(), ["attr", EqValid_Contain("att5"), ], {"_attr": None}),
        ]
    )
    def test__names(self, source, skip, _EXPECTED):
        Lambda(AttrAux_Existed(source).dump_dict, *skip).check_expected__assert(_EXPECTED)

    @pytest.mark.parametrize(
        argnames="cal_use, _EXPECTED",
        argvalues=[
            (EnumAdj_CallResolveStyle.DIRECT, (None, True, LAMBDA_TRUE, LAMBDA_RAISE,)),
            (EnumAdj_CallResolveStyle.EXC, (None, True, True, Exception,)),
            # (EnumAdj_CallResolveStyle.RAISE, Exception),          # need special tests!
            (EnumAdj_CallResolveStyle.RAISE_AS_NONE, (None, True, True, None,)),
            (EnumAdj_CallResolveStyle.BOOL, (False, True, True, False,)),
            (EnumAdj_CallResolveStyle.SKIP_CALLABLE, (None, True, None, None,)),
            (EnumAdj_CallResolveStyle.SKIP_RAISED, (None, True, True, None,)),
        ]
    )
    def test__callable_use(self, cal_use, _EXPECTED):
        result_dict = AttrAux_Existed(Victim).dump_dict(callables_resolve=cal_use)
        Lambda(dict.get, result_dict, "NONE").check_expected__assert(_EXPECTED[0])
        Lambda(dict.get, result_dict, "TRUE").check_expected__assert(_EXPECTED[1])
        Lambda(dict.get, result_dict, "LTRUE").check_expected__assert(_EXPECTED[2])
        Lambda(dict.get, result_dict, "RAISE").check_expected__assert(_EXPECTED[3])

    def test__callable_use__special_raise(self):
        try:
            result_dict = AttrAux_Existed(Victim).dump_dict(callables_resolve=EnumAdj_CallResolveStyle.RAISE)
        except:
            pass
        else:
            assert False


# =====================================================================================================================



from base_aux.base_values.m2_value_special import *


# =====================================================================================================================
class _Cls:
    def meth(self):
        pass
    @property
    def prop(self):
        return


def _explore():
    print(isinstance(_Cls.prop, property))  # True
    print(type(_Cls.prop))  # <class 'property'>
    print(type(_Cls().prop))  # <class 'NoneType'>


# =====================================================================================================================
OBJ_NAMES__BUILTINS = tuple(dir(globals().get("__builtins__")))
"""
('ArithmeticError', 'AssertionError', 'AttributeError', 'BaseException', 'BaseExceptionGroup', 'BlockingIOError', 'BrokenPipeError', 'BufferError', 'BytesWarning', 'ChildProcessError', 'ConnectionAbortedError', 'ConnectionError', 'ConnectionRefusedError', 'ConnectionResetError', 'DeprecationWarning', 'EOFError', 'Ellipsis', 'EncodingWarning', 'EnvironmentError', 'Exception', 'ExceptionGroup', 'False', 'FileExistsError', 'FileNotFoundError', 'FloatingPointError', 'FutureWarning', 'GeneratorExit', 'IOError', 'ImportError', 'ImportWarning', 'IndentationError', 'IndexError', 'InterruptedError', 'IsADirectoryError', 'KeyError', 'KeyboardInterrupt', 'LookupError', 'MemoryError', 'ModuleNotFoundError', 'NameError', 'None', 'NotADirectoryError', 'NotImplemented', 'NotImplementedError', 'OSError', 'OverflowError', 'PendingDeprecationWarning', 'PermissionError', 'ProcessLookupError', 'RecursionError', 'ReferenceError', 'ResourceWarning', 'RuntimeError', 'RuntimeWarning', 'StopAsyncIteration', 'StopIteration', 'SyntaxError', 'SyntaxWarning', 'SystemError', 'SystemExit', 'TabError', 'TimeoutError', 'True', 'TypeError', 'UnboundLocalError', 'UnicodeDecodeError', 'UnicodeEncodeError', 'UnicodeError', 'UnicodeTranslateError', 'UnicodeWarning', 'UserWarning', 'ValueError', 'Warning', 'WindowsError', 'ZeroDivisionError', '__build_class__', '__debug__', '__doc__', '__import__', '__loader__', '__name__', '__package__', '__spec__', 'abs', 'aiter', 'all', 'anext', 'any', 'ascii', 'bin', 'bool', 'breakpoint', 'bytearray', 'bytes', 'callable', 'chr', 'classmethod', 'compile', 'complex', 'copyright', 'credits', 'delattr', 'dict', 'dir', 'divmod', 'enumerate', 'eval', 'exec', 'exit', 'filter', 'float', 'format', 'frozenset', 'getattr', 'globals', 'hasattr', 'hash', 'help', 'hex', 'id', 'input', 'int', 'isinstance', 'issubclass', 'iter', 'len', 'license', 'list', 'locals', 'map', 'max', 'memoryview', 'min', 'next', 'object', 'oct', 'open', 'ord', 'pow', 'print', 'property', 'quit', 'range', 'repr', 'reversed', 'round', 'set', 'setattr', 'slice', 'sorted', 'staticmethod', 'str', 'sum', 'super', 'tuple', 'type', 'vars', 'zip')
"""


# COLLECTIONS =========================================================================================================
@final
class TYPES:
    """
    GOAL
    ----
    collect all types USEFUL variants
    using in isinstance (single or tuple)
    """
    # SINGLE ---------------------------
    NONE: type = type(None)
    NUMBER = int | float

    FUNCTION: type = type(lambda: True)
    METHOD: type = type(_Cls().meth)
    # PROPERTY: type    # cant check by type
    # MODULE: type    # cant check by type - only by method!

    # COLLECTIONS ---------------------------
    ELEMENTARY_SINGLE: tuple[type, ...] = (
        type(None),
        bool,
        int, float,
        str, bytes,
    )
    ELEMENTARY_COLLECTION: tuple[type, ...] = (
        tuple, list,
        set, dict,
    )
    ELEMENTARY: tuple[type, ...] = (
        *ELEMENTARY_SINGLE,
        *ELEMENTARY_COLLECTION,
    )


# =====================================================================================================================



from typing import *

from .tc0__base import *


# =====================================================================================================================
class TestCase(Base_TcAtcPtb):
    ATC_VOUT: int | None = 220
    PTB_SET_EXTON: bool = False
    PTB_SET_HVON: bool = True
    PTB_SET_PSON: bool = True

    _DESCRIPTION = "тест КЗ\nшины главного питания (MAIN)"

    # -----------------------------------------------------------------------------------------------------------------
    def run__wrapped(self) -> TYPING__RESULT_W_EXC:
        result_chain = ValidChains(
            chains=[
                ValidNoCum(
                    value_link=self.DEV_COLUMN.DUT.GET,
                    args__value="V12S",
                    validate_link=EqValid_LGTE_NumParsedSingle(ge=11, le=13).resolve,
                    name="GET+valid diapason",
                ),
                ValidFailContinue(
                    value_link=self.DEV_COLUMN.DUT.TEST,
                    args__value="SC12M",
                    kwargs__value={"__timeout": 40},
                    validate_link="PASS",
                    name="TEST",
                ),
                ValidSleep(1),
                ValidNoCum(
                    value_link=self.DEV_COLUMN.DUT.GET,
                    args__value="V12S",
                    validate_link=EqValid_LGTE_NumParsedSingle(ge=11, le=13).resolve,
                    name="GET+valid diapason",
                ),
            ]
        )
        return result_chain


# =====================================================================================================================



import datetime
import time
import pathlib

from base_aux.path1_dir.m1_dirpath import Resolve_DirPath
from base_aux.aux_argskwargs.m2_argskwargs_aux import *
from base_aux.base_enums.m2_enum1_adj import *


# =====================================================================================================================
# @final      # dont use final here! expect nesting for fileWork!???
class DirAux:   # NOTE: Init_SOURCE - dont nest!!!
    """
    GOAL
    ----
    collect all meths for directory include several files EXTERNAL work!
    single file work with FileAux class!
    """
    DIRPATH: TYPING.PATH_FINAL

    def __init__(self, dirpath: TYPING.PATH_DRAFT = None) -> None:
        self.DIRPATH = Resolve_DirPath(dirpath).resolve()

    # -----------------------------------------------------------------------------------------------------------------
    def check_exists(self, timeout: int = 0) -> bool:
        """
        NOTE
        ----
        any object - DIR/FILE!
        """
        for i in range(timeout):
            if self.DIRPATH.exists():
                return True
            time.sleep(1)

        return self.DIRPATH.exists()

    # -----------------------------------------------------------------------------------------------------------------
    def create_dirtree(self) -> bool:
        try:
            self.DIRPATH.mkdir(parents=True, exist_ok=True)
        except:
            pass

        if self.DIRPATH.exists():
            return True
        else:
            msg = f"[ERROR] CANT create {self.DIRPATH=}"
            print(msg)
            return False

    # -----------------------------------------------------------------------------------------------------------------
    def iter(
            self,
            *wmask: str,        # dont
            nested: bool = None,
            fsobj: EnumAdj_PathType = EnumAdj_PathType.ALL,
            str_names_only: bool = False,

            # time filter -----
            mtime: Union[None, datetime.datetime, datetime.timedelta] = None,   # acceptable for both FileAux/Dirs
            mtime_cmp: EnumAdj_CmpType = EnumAdj_CmpType.GE,
    ) -> Iterator[Union[pathlib.Path, str]] | NoReturn:
        """
        GOAL
        ----
        iter masked objects/names.
        """
        USE_DELTA = None
        if mtime is None:
            pass
        elif isinstance(mtime, datetime.datetime):
            mtime = mtime.timestamp()
        elif isinstance(mtime, datetime.timedelta):
            USE_DELTA = True
            pass
        elif not isinstance(mtime, (type(None), int, float)):
            raise TypeError(f"{mtime=}")

        wmask = wmask or ["*", ]
        # result = []

        for mask in wmask:
            mask = mask if not nested else f"**/{mask}"
            for path_obj in self.DIRPATH.glob(mask):
                if (
                        (fsobj == EnumAdj_PathType.FILE and path_obj.is_file())
                        or
                        (fsobj == EnumAdj_PathType.DIR and path_obj.is_dir())
                        or
                        fsobj == EnumAdj_PathType.ALL
                ):
                    if mtime:
                        mtime_i = path_obj.stat().st_mtime
                        if USE_DELTA:
                            mtime_i = datetime.datetime.now().timestamp() - mtime_i

                        if (
                                mtime_cmp == EnumAdj_CmpType.LE and not mtime_i <= mtime  # OLDER
                                or
                                mtime_cmp == EnumAdj_CmpType.LT and not mtime_i < mtime
                                or
                                mtime_cmp == EnumAdj_CmpType.GE and not mtime_i >= mtime  # NEWER
                                or
                                mtime_cmp == EnumAdj_CmpType.GT and not mtime_i > mtime
                        ):
                            continue

                    if str_names_only:
                        result_i = path_obj.name
                    else:
                        result_i = path_obj

                    yield result_i

        # print(f"{result=}")
        # return result

    def iter_files(self, *wmask, **kwargs) -> Iterator[Union[pathlib.Path, str]]:
        yield from self.iter(*wmask, fsobj=EnumAdj_PathType.FILE, **kwargs)

    def iter_dirs(self, *wmask, **kwargs) -> Iterator[Union[pathlib.Path, str]]:
        yield from self.iter(*wmask, fsobj=EnumAdj_PathType.DIR, **kwargs)

    # -----------------------------------------------------------------------------------------------------------------
    def delete_blank(self, raise_fails: bool = None) -> bool | NoReturn:
        """
        GOAL
        ----
        delete SELF directory if blank!
        """
        result = True
        try:
            self.DIRPATH.rmdir()
        except Exception as exc:  # TODO: separate AccessPermition/FilesExists
            result = False
            if raise_fails:
                raise exc
        return result

    def delete_blank_items_wmask(
            self,
            *wmask: str,
            nested: bool = None,
    ) -> bool | NoReturn:
        """
        GOAL
        ----
        delete INTERNAL dirs in SELF if blanks!
        """
        # TODO: NOT WORKING!!!!! FINISH!!!! cant delete by access reason!!!
        result = True
        for dirpath in self.iter_dirs(*wmask, nested=nested):
            result &= DirAux(dirpath).delete_blank()

        return result

    def delete_dirtree(self, raise_fails: bool = None) -> bool | NoReturn:
        return self.delete_items(self.DIRPATH, raise_fails=raise_fails)

    @classmethod
    def delete_items(cls, *paths: TYPING.PATH_FINAL, raise_fails: bool = None) -> bool | NoReturn:
        result = True
        for path in paths:

            if path.is_file():
                try:
                    path.unlink()
                except Exception as exc:
                    result = False
                    if raise_fails:
                        raise exc

            elif path.is_dir():
                result &= cls.delete_items(*cls(path).iter_files(), raise_fails=raise_fails)
                result &= cls.delete_items(*cls(path).iter_dirs(), raise_fails=raise_fails)
                result &= DirAux(path).delete_blank(raise_fails=raise_fails)

            # TODO: if link

        return result


# =====================================================================================================================



from typing import *

from PyQt5.QtGui import *
from PyQt5.QtCore import *
from PyQt5.QtWidgets import *

from base_aux.base_nest_dunders.m5_iter_annots_values import NestIter_AnnotValues


# =====================================================================================================================
def format_make(color_fg: Any = None, style: str = '', color_bg: Any = None) -> QTextCharFormat:
    """
     :returns: QTextCharFormat с указанными атрибутами указанными текстом
     """
    result = QTextCharFormat()

    if color_fg:
        color_fg = QColor(color_fg)
        # color_fg.setNamedColor(color_fg_name)
        result.setForeground(color_fg)

    if color_bg:
        color_bg = QColor(color_bg)
        # color_bg.setNamedColor(color_bg_name)
        result.setBackground(color_bg)

    if 'bold' in style:
        result.setFontWeight(QFont.Bold)
    if 'italic' in style:
        result.setFontItalic(True)

    return result


# =====================================================================================================================
class HlStyle(NamedTuple):
    FORMAT: QTextCharFormat
    P_ITEMS: list[str]
    P_TEMPLATES: list[str] = [r'%s', ]  # FOR ONE LINE!!!
    INDEX: int = 0

    def get_patterns(self) -> set[str]:
        if not self.P_ITEMS:
            result = [pattern for pattern in self.P_TEMPLATES]

        elif not self.P_TEMPLATES:
            result = [pattern for pattern in self.P_ITEMS]

        else:
            result = []
            for template in self.P_TEMPLATES:
                result += [template % item for item in self.P_ITEMS]

        return set(result)

    def get_regexps(self) -> list[QRegExp]:
        return [QRegExp(pattern, cs=Qt.CaseSensitivity(0)) for pattern in self.get_patterns()]

    def get_rules(self) -> list[tuple[QRegExp, int, QTextCharFormat]]:
        return [(regexp, self.INDEX, self.FORMAT) for regexp in self.get_regexps()]


# =====================================================================================================================
class HlStyles(NestIter_AnnotValues):
    """
    NOTE
    ----
    ORDER is important!
    LAST rule - would be final!

    EXAMPLE
    -------
    see HlStylesPython
    """
    def get_rules(self) -> list[tuple[QRegExp, int, QTextCharFormat]]:
        result = []
        for group in self:
            result.extend(group.get_rules())
        return result


# =====================================================================================================================
class HlStylesPython(HlStyles):
    KEYWORD: HlStyle = HlStyle(
        FORMAT=format_make('blue'),
        P_ITEMS=[
            'assert', 'exec',
            'global', 'import',
            'lambda', 'print', 'del',

            # def
            'class', "def",

            # blocks
            'if', 'for', 'while', 'try', 'from',
            'elif', 'else', 'except', 'continue', 'finally',
            'raise', 'return', 'yield', 'break', "pass",

            # operators word
            'is', 'or', 'and', 'in', 'not',

            # types
            "int", "bool",  # add finish

            # values
            'None', 'True', 'False',
        ],
        P_TEMPLATES=[
            r'\b%s\b',
        ],
    )
    OPERATOR_SIGN: HlStyle = HlStyle(
        FORMAT=format_make('red'),
        P_ITEMS=[
            '=',
            # Comparison
            '==', '!=', '<', '<=', '>', '>=',
            # Arithmetic
            r'\+', '-', r'\*', '/', '//', r'\%', r'\*\*',
            # In-place
            r'\+=', '-=', r'\*=', '/=', r'\%=',
            # Bitwise
            r'\^', r'\|', r'\&', r'\~', '>>', '<<',
        ],
        P_TEMPLATES=[
        ],
    )
    BRACE: HlStyle = HlStyle(
        FORMAT=format_make('darkGray'),
        P_ITEMS=[
            r'\{', r'\}', r'\(', r'\)', r'\[', r'\]',
        ],
        P_TEMPLATES=[
        ],
    )
    DEF: HlStyle = HlStyle(
        FORMAT=format_make('black', 'bold'),
        P_ITEMS=[
            "def", "class"
        ],
        P_TEMPLATES=[
            r'\b%s\b\s*(\w+)',
        ],
        INDEX=1,
    )
    SELF: HlStyle = HlStyle(
        FORMAT=format_make('black', 'italic'),
        P_ITEMS=[
        ],
        P_TEMPLATES=[
            r'\bself\b',
        ],
    )
    NUMBERS: HlStyle = HlStyle(
        FORMAT=format_make('brown'),
        P_ITEMS=[
        ],
        P_TEMPLATES=[
            r'\b[+-]?[0-9]+[lL]?\b',
            r'\b[+-]?0[xX][0-9A-Fa-f]+[lL]?\b',
            r'\b[+-]?[0-9]+(?:\.[0-9]+)?(?:[eE][+-]?[0-9]+)?\b',
        ],
    )
    COMMENT: HlStyle = HlStyle(
        FORMAT=format_make('darkGreen', 'italic'),
        P_ITEMS=[
        ],
        P_TEMPLATES=[
            # r'#[^\n]*',
            r'#.*',
        ],
    )
    STRING: HlStyle = HlStyle(
        FORMAT=format_make('magenta'),
        P_ITEMS=[
        ],
        P_TEMPLATES=[
            r'"[^"\\]*(\\.[^"\\]*)*"',
            r"'[^'\\]*(\\.[^'\\]*)*'",
        ],
    )


class HlStylesMultiline(HlStyles):
    STRING_3S: HlStyle = HlStyle(
        FORMAT=format_make('darkMagenta'),
        P_ITEMS=[
        ],
        P_TEMPLATES=[
            "'''",
        ],
        INDEX=1,
    )
    STRING_3D: HlStyle = HlStyle(
        FORMAT=format_make('darkMagenta'),
        P_ITEMS=[
        ],
        P_TEMPLATES=[
            '"""',
        ],
        INDEX=2,
    )


# ---------------------------------------------------------------------------------------------------------------------
class HlStylesExample(HlStylesPython):
    RESULT_TRUE: HlStyle = HlStyle(
        FORMAT=format_make("", "", "lightGreen"),
        P_ITEMS=[
            "True"
        ],
        P_TEMPLATES=[
            r'.*=\s*%s.*',
        ],
    )
    RESULT_FALSE: HlStyle = HlStyle(
        FORMAT=format_make("", "", "pink"),
        P_ITEMS=[
            "False"
        ],
        P_TEMPLATES=[
            r'.*=\s*%s.*',
        ],
    )


# =====================================================================================================================
class Highlighter(QSyntaxHighlighter):
    # settings --------------------
    STYLES_LINE: HlStyles = HlStyles()
    STYLES_MULTILINE: HlStyles = HlStyles()

    # aux --------------------
    RULES_LINE: list[tuple[QRegExp, int, QTextCharFormat]] = []
    RULES_MULTILINE: list[tuple[QRegExp, int, QTextCharFormat]] = []

    def __init__(self, document: QTextDocument, styles: HlStyles = None, styles_multiline: HlStyles = None):
        super().__init__(document)
        if styles:
            self.STYLES_LINE = styles
        if styles_multiline:
            self.STYLES_MULTILINE = styles_multiline

        self.RULES_LINE = self.STYLES_LINE.get_rules()
        self.RULES_MULTILINE = self.STYLES_MULTILINE.get_rules()

    def highlightBlock(self, text, *args) -> None:
        """Применить выделение синтаксиса к данному блоку текста. """
        for expression, nth, format in self.RULES_LINE:
            index = expression.indexIn(text, 0)
            while index >= 0:
                index = expression.pos(nth)
                length = len(expression.cap(nth))
                self.setFormat(index, length, format)
                index = expression.indexIn(text, index + length)

        # self.setCurrentBlockState(0)
        self.apply_multiline(text)

    def apply_multiline(self, text):
        for delimiter, in_state, style in self.RULES_MULTILINE:
            if self.previousBlockState() == in_state:
                start = 0
                add = 0
            else:
                start = delimiter.indexIn(text)
                add = delimiter.matchedLength()

            while start >= 0:
                end = delimiter.indexIn(text, start + add)
                if end >= add:
                    length = end - start + add + delimiter.matchedLength()
                    self.setCurrentBlockState(0)
                else:
                    self.setCurrentBlockState(in_state)
                    length = len(text) - start + add

                self.setFormat(start, length, style)
                start = delimiter.indexIn(text, start + length)

            if self.currentBlockState() == in_state:
                return True
            else:
                return False


# =====================================================================================================================
EXAMPLE_TEXT = """
from typing import *
# привет hello 123
class Cls:  # привет hello 123
    attr_dict: dict = {1: "123"}
    attr_tuple: tuple[Any] = (None, "hello123 123", 'hello123 123')
    attr_list: list[Any] = [0, "1", '2', '''string''', True, int]
    def meth(self) -> bool:
        pass

print(Cls.__annotations__)
print(Cls().__annotations__)
{'attr1': <class 'int'>}

# USER ----- =========
result=True123
result=False 123

# COLLECTIONS ----- =========
()[]{} ([{

[1,2, 3]
[
    1,
    2,
]

(1,2, 3)
(
    1,
    2,
)

{1,2, 3}
{
    1,
    2,
}

{1: "1", 2: "2"}
{
    1: "1":,
    2: "2"
}

# TEXT ----- =========
'''
'''

'''
def hello():
    pass

"""


def start_example():
    app = QApplication([])
    PTE = QPlainTextEdit()

    # font = QFont()
    # font.setPointSize(12)
    # PTE.setFont(font)

    highlight = Highlighter(document=PTE.document(), styles=HlStylesExample(), styles_multiline=HlStylesMultiline())   # need to keep in not used var!
    PTE.show()
    PTE.setPlainText(EXAMPLE_TEXT)
    print(f"{PTE.document()=}")
    # ObjectInfo(PTE.document()).print()

    app.exec_()


# =====================================================================================================================
if __name__ == '__main__':
    start_example()


# =====================================================================================================================



from base_aux.aux_dict.m4_dict_diff import *
from base_aux.aux_attr.m4_dump1_dumping1_dict import *
from base_aux.base_types.m0_static_typing import *


# =====================================================================================================================
class Base_AttrDiff(Base_DiffResolve):
    """
    GOAL
    ----
    cmp several objects by attr values.
    MAIN GOAL - cmp several states of one obj after DumpAttrObj.
        but it seems cant be realisable! cause we cant use at same time/in one code line several states for one object!
    so
        1/ for one object - use state dumps with DictDiff
        2/ for several identical objects - use this staff! AttrDiff

    SPECIALLY CREATED FOR
    ---------------------
    """
    ARGS: tuple[Any, ...]
    DIFF: TYPING.DICT_STR_TUPLE_ANY
    __diff: TYPING.DICT_STR_TUPLE_ANY

    CLS_ATTR_DUMPING: type[Base_AttrDictDumping] = AttrDictDumping_Existed

    def resolve(self) -> TYPING.DICT_STR_TUPLE_ANY:
        DICTS = []
        for obj in self.ARGS:
            dict_i = self.CLS_ATTR_DUMPING(obj).resolve()
            DICTS.append(dict_i)

        self.__diff = DictDiff(*DICTS).resolve()
        return self.__diff


# =====================================================================================================================
class AttrDiff_Existed(Base_AttrDiff):
    CLS_ATTR_DUMPING: type[Base_AttrDictDumping] = AttrDictDumping_Existed


class AttrDiff_AnnotsAll(Base_AttrDiff):
    CLS_ATTR_DUMPING: type[Base_AttrDictDumping] = AttrDictDumping_AnnotsAll


class AttrDiff_AnnotsLast(Base_AttrDiff):
    CLS_ATTR_DUMPING: type[Base_AttrDictDumping] = AttrDictDumping_AnnotsLast


# =====================================================================================================================
if __name__ == "__main__":
    pass


# =====================================================================================================================



from typing import *

from base_aux.base_values.m3_exceptions import *


# =====================================================================================================================
ARGS_FINAL__BLANK = ()
KWARGS_FINAL__BLANK = {}


# =====================================================================================================================
class Mark_ValueSpecial:
    """
    GOAL
    ----
    BASE for result values like class (dont use instances!)
    using exact values wich can be directly compared and using as special universal values/states.
    values with Special meaning!

    NOTE
    ----
    never instantiate it! use value only as Class!

    SPECIALLY CREATED FOR
    ---------------------
    DictDiff instead of Enum values! - in there we have __Eq which cause incorrect usage/results!
    """

    # TODO: add Meta EqCls with Exc??? to cmp with exact values??? - no! use special logic when do cmp!
    def __init__(self) -> NoReturn:
        msg = f"Mark_ValueSpecial NEVER INITTIATE! use direct CLASS!"
        raise Exc__WrongUsage(msg)

    # def __bool__(self):     # CANT USE!!! it works only on instance!!!
    #     return False

    # TODO: add classmethod! - not working!!!
    # def __str__(self):
    #     return f"{self.__class__.__name__}"


# =====================================================================================================================
class NoValue(Mark_ValueSpecial):
    """
    DOUPTS
    ------
    1. DEPRECATE???=====NO! used in valid/value
    2. use direct ArgsEmpty???/ArgsKwargs()??

    GOAL
    ----
    it is different from Default!
    there is no value!
    used when we need to change logic with not passed value!

    SPECIALLY CREATED FOR
    ---------------------
    Valid as universal validation object under cmp other base_types!

    USAGE
    -----
    class Cls:
        def __init__(self, value: Any | type[NoValue] | NoValue = NoValue):
            self.value = value

        def __eq__(self, other):
            if self.value is NoValue:
                return other is True
                # or
                return self.__class__(other).run()
            else:
                return other == self.value

        def run(self):
            return bool(self.value)
    """


# =====================================================================================================================
class Raised(Mark_ValueSpecial):
    """
    GOAL
    ----
    just a mirror for GA_NotExists
    """


# =====================================================================================================================
class Skipped(Mark_ValueSpecial):
    """
    GOAL
    ----
    create special value for result in Lambda.resolve__style when skipped
    """


# =====================================================================================================================
class _ValueSpecial:
    """
    GOAL
    ----
    just a collection for special values!

    USE instance!
    """
    NOVALUE: type[Mark_ValueSpecial] = NoValue
    RAISED: type[Mark_ValueSpecial] = Raised
    SKIPPED: type[Mark_ValueSpecial] = Skipped

    def __iter__(self) -> Iterable[type]:
        """
        GOAL
        ----
        iter values
        """
        yield from self.__annotations__.values()

    def __contains__(self, item: Any) -> bool:
        """
        GOAL
        ----
        check value is special!
        """
        return item in self


# =====================================================================================================================
VALUE_SPECIAL = _ValueSpecial()


# =====================================================================================================================



from typing import *

from enum import Enum
from base_aux.base_values.m2_value_special import NoValue


# =====================================================================================================================
# TODO: make own class! EnumIc with metaclass!
class NestEq_EnumAdj(Enum):
    """
    # NOTE
    # ----
    # DEL=work IC only on Eq! not working with Contains and Init!!! need edit Metaclass!

    CONSTRAINTS
    -----------
    1/ USE EqEnum: ONLY at SETTINGS values! and params values! so it is PARAMS using!!!
    2/ DONT USE: with result values! (for this case see Mark_ValueSpecial specially created!)

    GOAL
    ----
    add user friendly cmp objects with final values

    for std object it is False but here its is correct!
    assert Enum(1) != 1
    assert NestEq_EnumAdj(1) == 1
    """
    # FIXME: cant create! - not work!!!
    # def __new__(cls, value_druft: Any):
    #     value_final = cls._value__get_original(value_druft)
    #     return cls(value_final)

    @classmethod
    def _value__get_original(cls, value_druft: Any) -> Any | NoValue:
        for item in cls:
            value_original = item.value
            if isinstance(value_druft, str):
                if value_druft.lower() == str(value_original).lower():
                    return value_original
            if value_druft == value_original:
                return value_original

        return NoValue

    # def __eq__(self, other) -> bool:
    #     if isinstance(other, self.__class__):
    #         return self.value == other.value        # or str(self.value).lower() == str(other.value).lower()    # NO!!!
    #
    #     else:
    #         for enum_i in self.__class__:
    #             if isinstance(other, str):
    #                 if str(enum_i.value).lower() == str(other).lower():
    #                     return True
    #             else:
    #                 try:
    #                     if enum_i.value == other or other == enum_i.value:
    #                         return True
    #                 except:
    #                     pass
    #     return False

    def __eq__(self, other) -> bool:
        # OBVIOUS -----------
        if isinstance(other, self.__class__):
            return self.value == other.value

        # CMP AVAILABLE -----
        cmp_available = False

        try:
            if other in self.__class__:
                cmp_available = True
        except:
            pass

        try:
            if other in self:
                cmp_available = True
        except:
            pass

        if cmp_available:
            return other == self.value or self == self.__class__(other)
        else:
            return False

    # TODO: add Contain classmeth???  cant understand! need metaclass!
    # @classmethod
    # def __contains__(cls, other) -> bool:
    #     if isinstance(other, cls):
    #         other = other.value
    #
    #     for enum_i in cls:
    #         if enum_i.value == other or str(enum_i.value).lower() == str(other).lower():
    #             return True
    #     else:
    #         return False

    @classmethod
    def as_dict__all(cls):
        """
        GOAL
        ----
        get all existed items in class
        """
        return {member.name: member.value for member in cls}


# =====================================================================================================================



from typing import *
import pytest

from base_aux.aux_attr.m4_kits import *


# =====================================================================================================================
def test__values():
    class Victim(Base_AttrKit):
        A1: Any
        A2: Any = None
        A3 = None
        DICT: dict = {}

    try:
        Victim()
    except:
        pass
    else:
        assert False

    assert Victim(a1=1).A1 == 1
    assert Victim(A1=1).a1 == 1
    assert Victim(1).a1 == 1
    assert Victim(1).A1 == 1

    assert Victim(1, a1=2).A1 == 2

    assert Victim(1).A1 == 1
    assert Victim(1).A2 == None
    assert Victim(1).A3 == None

    assert Victim(1, 1, 1).A1 == 1
    assert Victim(1, 1, 1).A2 == 1
    assert Victim(1, 1, 1).A3 == None
    assert Victim(1, 1, a3=1).A3 == 1

    # mutable
    victim = Victim(1, 1, a3=1)
    assert victim.DICT == Victim.DICT
    assert victim.DICT is not Victim.DICT

    victim.DICT[1]=1
    assert victim.DICT[1] == 1
    assert victim.DICT != Victim.DICT


# ---------------------------------------------------------------------------------------------------------------------
def test__eq():
    class Victim:
        A0: Any
        A1: Any = 1

    assert AttrKit_Blank(a1=1) != Victim()
    assert AttrKit_Blank(A1=1) == Victim()
    assert AttrKit_Blank(a1=11) != Victim()
    assert AttrKit_Blank(a0=1) != Victim()

    try:
        AttrKit_AuthTgBot(1)
    except:
        pass
    else:
        assert False

    assert AttrKit_AuthTgBot(1, 2, 3).token == 3


# =====================================================================================================================
def test__cls_name():
    assert NestInit_AnnotsAttr_ByArgsKwargs().__class__.__name__ == f"NestInit_AnnotsAttr_ByArgsKwargs"

    class Victim(NestInit_AnnotsAttr_ByArgsKwargs):
        A1: Any = None

    assert Victim().__class__.__name__ == f"Victim"


# =====================================================================================================================



from typing import *


# =====================================================================================================================
class Meta_GaClsPrefix_Example(type):
    """
    NOTE
    ----
    THIS IS ONLY AS EXAMPLE!!! in any situation you need to create specific Meta-class!!!

    GOAL
    ----
    get attributes from CLASS! not instance!

    CREATED SPECIALLY FOR
    ---------------------
    the main goal

    BEST USAGE
    ----------
    for example usages see valid_getter.Base_ReqCheckStr, annot_attrs. or here _GetattrCls_Meta__Echo
    """
    # TODO: add example! + tests! +REF!
    # dont change markers! use exists!
    # FIXME: change exact markers not direct but as LIST!!!
    _MARKER__BOOL_IF: str = "bool_if__"
    _MARKER__BOOL_IF_NOT: str = "bool_if_not__"
    _MARKER__RAISE_IF: str = "raise_if__"
    _MARKER__RAISE_IF_NOT: str = "raise_if_not__"

    def __getattr__(cls, item: str):
        """if no exists attr/meth
        """
        if item.lower().startswith(cls._MARKER__BOOL_IF.lower()):
            attr_name = item.lower().replace(cls._MARKER__BOOL_IF.lower(), "")
            return lambda: cls.check(values=attr_name, _raise=False, _reverse=False, _meet_true=False)
        elif item.lower().startswith(cls._MARKER__BOOL_IF_NOT.lower()):
            attr_name = item.lower().replace(cls._MARKER__BOOL_IF_NOT.lower(), "")
            return lambda: cls.check(values=attr_name, _raise=False, _reverse=True, _meet_true=False)

        elif item.lower().startswith(cls._MARKER__RAISE_IF.lower()):
            attr_name = item.lower().replace(cls._MARKER__RAISE_IF.lower(), "")
            return lambda: not cls.check(values=attr_name, _raise=True, _reverse=True, _meet_true=False) or None
        elif item.lower().startswith(cls._MARKER__RAISE_IF_NOT.lower()):
            attr_name = item.lower().replace(cls._MARKER__RAISE_IF_NOT.lower(), "")
            return lambda: not cls.check(values=attr_name, _raise=True, _reverse=False, _meet_true=False) or None

        else:
            msg = f"[ERROR] META:'{cls.__name__}' CLASS has no attribute '{item}'"
            raise AttributeError(msg)


# =====================================================================================================================



from typing import *

from base_aux.testplans.stand import *
from base_aux.base_nest_dunders.m5_iter_annots_values import NestIter_AnnotValues

from .DEVICES import dev_lines
from .Example import (
    tc1_direct,
    tc2_reverse,
    tc3_atc,
)
from .psu800 import (
    tc1_none_1_exist_psu,
    tc1_none_2_test_gnd,
    tc1_none_3_off,
    tc1_none_4_on,

    tc2_ext_1_test_pmbus,
    tc2_ext_2_off,
    tc2_ext_3_on,

    tc3_hv_1_off,
    tc3_hv_2_on,

    tc4_hv_1_test_SC12S,
    tc4_hv_2_test_SC12M,
    tc4_hv_3_test_LD12S,
    tc4_hv_4_test_LOAD,

    tc8_hv1_150_off,
    tc8_hv2_160_on,
    tc8_hv3_250_on,
    tc8_hv4_260_off,
)


# =====================================================================================================================
class Tp_Example(Base_Stand):
    NAME = "[пример] с пустыми устройствами"
    DEV_LINES = dev_lines.DeviceLines__AtcPtbDummy()
    TCSc_LINE = TableLine(
        tc1_direct.TestCase,
        tc2_reverse.TestCase,
    )


# =====================================================================================================================
class Tp_Example2(Base_Stand):
    NAME = "[пример] с реальными устройствами"
    DEV_LINES = dev_lines.DeviceLines__Psu800()
    TCSc_LINE = TableLine(
        tc1_direct.TestCase,
        tc2_reverse.TestCase,
        tc3_atc.TestCase,
    )


# =====================================================================================================================
class Tp_Psu800(Base_Stand):
    NAME = "[ОТК] БП800"
    DEV_LINES = dev_lines.DeviceLines__Psu800()
    TCSc_LINE = TableLine(
        tc1_none_1_exist_psu.TestCase,
        tc1_none_2_test_gnd.TestCase,
        tc1_none_3_off.TestCase,
        tc1_none_4_on.TestCase,

        tc2_ext_1_test_pmbus.TestCase,
        tc2_ext_2_off.TestCase,
        tc2_ext_3_on.TestCase,

        tc3_hv_1_off.TestCase,
        tc3_hv_2_on.TestCase,

        tc4_hv_1_test_SC12S.TestCase,
        tc4_hv_2_test_SC12M.TestCase,
        tc4_hv_3_test_LD12S.TestCase,
        tc4_hv_4_test_LOAD.TestCase,

        tc8_hv1_150_off.TestCase,
        tc8_hv2_160_on.TestCase,
        tc8_hv3_250_on.TestCase,
        tc8_hv4_260_off.TestCase,
    )


# =====================================================================================================================
class Stands(NestIter_AnnotValues):
    TP_EXAMPLE: Base_Stand = Tp_Example()
    TP_EXAMPLE2: Base_Stand = Tp_Example2()
    TP_PSU800: Base_Stand = Tp_Psu800()


# =====================================================================================================================



from base_aux.versions.m3_derivatives import *
from base_aux.testplans.tp_manager import *
from base_aux.servers.m1_client_requests import *

from TESTPLANS.stands import Stands


# =====================================================================================================================
Version_Python().raise_if_not__check_ge("3.11", COMMENT="need greater equal!")


# =====================================================================================================================
class Client_RequestItem_Tp(Client_RequestItem):
    LOG_ENABLE = True

    RETRY_LIMIT = 1
    RETRY_TIMEOUT = 1

    HOST: str = "192.168.74.20"
    PORT: int = 8080
    ROUTE: str = "results"

    SUCCESS_IF_FAIL_CODE = True


class Client_RequestsStack_Tp(Client_RequestsStack):
    LOG_ENABLE = True
    REQUEST_CLS = Client_RequestItem_Tp


# =====================================================================================================================
class TpManager__Example(TpManager):
    LOG_ENABLE = True

    STANDS = Stands
    STAND = Stands.TP_PSU800

    API_SERVER__CLS = TpApi_FastApi
    api_client: Client_RequestsStack = Client_RequestsStack_Tp()  # FIXME: need fix post__results!!!!
    # api_client: Client_RequestsStack = None

    GUI__START = True
    API_SERVER__START = True

    # def post__tc_results(self, tc_inst: Base_TestCase) -> None:
    #     # CHECK ------------------------------------------
    #     if not self.api_client or tc_inst.result is None:
    #         return
    #
    #     # WORK ------------------------------------------
    #     # TODO: need CREATE good Model + generate it + use pydantic in MW + validate + add correct in SQL
    #     body = {
    #         "sn": tc_inst.DEVICES__BREEDER_INST.DUT.SN,
    #         "timestamp": tc_inst.timestamp_last,
    #         "factory_record": {
    #             "name": "name_stand",
    #             "stand_number": 111,
    #             "shift_number": 222,
    #         },
    #         "test_result": "PASS" if tc_inst.result else "FAIL",
    #         "components": {
    #             "component1": "comp1",
    #             "component2": "comp2",
    #         },
    #         "versions": {
    #             "board_version": "ver_board",
    #             "firmware_version": "ver_fir",
    #         },
    #         "test_log": {
    #             "test_name": "test1",
    #             "test_result": "PASS" if tc_inst.result else "FAIL",
    #             "parameters": {
    #                 "param1": 111,
    #                 "param2": 222,
    #             },
    #             "log_file": "log_file.txt",
    #         },
    #         "history_record": {
    #             "status": "Shipped",
    #             "other_databases": {
    #                 "database1": "DB1",
    #                 "database2": "DB2",
    #             },
    #         },
    #
    #         # **tc_inst.get__results().dict(),
    #     }
    #     print(body)
    #     self.api_client.send(body=body)


# =====================================================================================================================
class TpInsideApi_Runner__example(TpInsideApi_Runner):
    TP_CLS = TpManager__Example


# =====================================================================================================================
def run_direct():
    TpManager__Example()


def run_over_api():
    TpInsideApi_Runner__example()


# =====================================================================================================================
if __name__ == "__main__":
    run_direct()
    # run_over_api()


# =====================================================================================================================



from base_aux.base_nest_dunders.m1_init0_reinit2_lambdas_resolve import NestInit_AttrsLambdaResolve
from base_aux.privates.m1_privates import *
from base_aux.aux_attr.m4_kits import *
from base_aux.alerts.m1_alert0_base import *

import telebot


# =====================================================================================================================
class RecipientTgID(Base_AttrKit):
    MyTgID: str


# =====================================================================================================================
class AlertTelegram(Base_Alert):
    """realisation for sending Telegram msg
    """
    # SETTINGS ------------------------------------
    CONN_ADDRESS: AttrKit_AuthNamePwd = PvLoaderIni_AuthTgBot(keypath=("TGBOT_DEF", ))
    RECIPIENT: AttrKit_AuthNamePwd = PvLoaderIni(target=RecipientTgID, keypath=("TG_ID",))

    # AUX -----------------------------------------
    _conn: telebot.TeleBot

    def _connect_unsafe(self) -> Union[bool, NoReturn]:
        self._conn = telebot.TeleBot(token=self.CONN_ADDRESS.TOKEN)
        return True

    def _send_unsafe(self) -> Union[bool, NoReturn]:
        self._conn.send_message(chat_id=self.RECIPIENT.MyTgID, text=self._msg_compose())
        return True


# =====================================================================================================================
if __name__ == "__main__":
    victim = AlertTelegram()
    victim.send_msg("hello")
    victim.send_msg("hello2")
    victim.wait()
    victim.send_msg("hello3")
    victim.wait()


# =====================================================================================================================



import pytest

from base_aux.base_lambdas.m1_lambda import *
from base_aux.base_values.m4_primitives import *

from base_aux.aux_eq.m2_eq_aux import *


# =====================================================================================================================
@pytest.mark.parametrize(
    argnames="source, args, _EXPECTED",
    argvalues=[
        (1, 1,              (True, True, False,     True, True, False)),
        (1, 2,              (False, False, True,    False, False, True)),
        (LAMBDA_TRUE, True, (False, False, True,    False, False, True)),

        # __EQ__=in single obj!
        (ClsEq(1), 1,       (True, True, False,     True, True, False)),
        (ClsEq(1), 2,       (False, False, True,    False, False, True)),
        (1, ClsEq(1),       (True, True, False,    True, True, False)),
        (2, ClsEq(1),       (False, False, True,    False, False, True)),

        (INST_EQ_RAISE, 1,  (Exception, False, True,    Exception, False, True)),
        (1, INST_EQ_RAISE,  (Exception, False, True,    Exception, False, True)),

        # __EQ__=in both objs!
        (INST_EQ_TRUE, INST_EQ_FALSE, (True, True, False,   True, True, False)),
        (INST_EQ_FALSE, INST_EQ_TRUE, (False, False, True,  True, True, False)),

    ]
)
def test__eq_aux(source, args, _EXPECTED):
    Lambda(EqAux(source).check_oneside__exc, args).check_expected__assert(_EXPECTED[0])
    Lambda(EqAux(source).check_oneside__bool, args).check_expected__assert(_EXPECTED[1])
    Lambda(EqAux(source).check_oneside__reverse, args).check_expected__assert(_EXPECTED[2])

    Lambda(EqAux(source).check_doubleside__exc, args).check_expected__assert(_EXPECTED[3])
    Lambda(EqAux(source).check_doubleside__bool, args).check_expected__assert(_EXPECTED[4])
    Lambda(EqAux(source).check_doubleside__reverse, args).check_expected__assert(_EXPECTED[5])


# =====================================================================================================================



import pytest
from base_aux.breeders.m2_table_inst import *
from base_aux.testplans.devices_base import *
from base_aux.testplans.devices_kit import DeviceKit


# =====================================================================================================================
class Test__DeviceBase:
    @classmethod
    def setup_class(cls):
        pass
        cls.Victim = type("Victim", (Base_Device,), {})

    # @classmethod
    # def teardown_class(cls):
    #     pass
    #
    # def setup_method(self, method):
    #     pass
    #
    # def teardown_method(self, method):
    #     pass

    # -----------------------------------------------------------------------------------------------------------------
    def test__1(self):
        victim = self.Victim()

    def test__2(self):
        victim = self.Victim()
        assert victim.INDEX is None

        # victim = self.Victim(2)
        # assert victim.INDEX == 2


# =====================================================================================================================
@pytest.mark.skip    # FIXME: ref or not!
class Test__DevicesLines:
    @classmethod
    def setup_class(cls):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    def setup_method(self, method):
        pass
        self.Victim: DeviceKit = type("Victim", (DeviceKit,), {})

    def teardown_method(self, method):
        pass

    # -----------------------------------------------------------------------------------------------------------------
    def test__DUT_COUNT(self):
        # 1 -----------------------------------------------------
        self.Victim.COUNT = 1

        victim = self.Victim(DUT=TableLine(0))
        assert victim.DUT == self.Victim.DUT[0]

        # 2 ------------------------------------------------------
        self.Victim.COUNT = 2
        #
        # # INSTANCE ----------------------
        victim = self.Victim(0)
        assert victim.DUT.INDEX == 0
        assert victim.DUT == self.Victim.DUT[0]
        assert victim.DUT == victim.DUT[0]

        victim = self.Victim(1)
        assert victim.DUT.INDEX == 1
        assert victim.DUT == self.Victim.DUT[1]
        assert victim.DUT == victim.DUT[1]

        assert victim.DUT[0] != victim.DUT[1]
        assert self.Victim.DUT[0] != self.Victim.DUT[1]

    # -----------------------------------------------------------------------------------------------------------------
    def test__CLS_SINGLE__CLS(self):
        self.Victim.COUNT = 2
        self.Victim.ATC = Base_Device

        assert hasattr(self.Victim, "DUT") is True
        assert hasattr(self.Victim, "ATC") is False

        assert self.Victim["DUT"] is not None
        assert self.Victim["ATC"] is not None
        assert self.Victim["PTB"] is not None

        # DISCONNECT
        self.Victim.disconnect()

    def test__CLS_SINGLE__INSTANCE(self):
        self.Victim.COUNT = 2
        self.Victim.ATC = Base_Device

        victim = self.Victim(1)    # FIXME: IS IN BROKEN?????

        assert victim.DUT == victim.DUT[1]
        try:
            victim.PTB
        except:
            pass
        else:
            assert False

        assert hasattr(victim, "DUT") is True
        try:
            hasattr(victim, "ATC")
        except:
            pass
        else:
            assert False

        assert victim["DUT"] is not None
        assert victim["DUT"] is not None
        assert victim["PTB"] is not None

    # -----------------------------------------------------------------------------------------------------------------
    def test__CLS(self):
        self.Victim.COUNT = 2
        self.Victim.PTB = Base_Device

        assert hasattr(self.Victim, "DUT") is True
        assert hasattr(self.Victim, "PTB") is True

        assert len(self.Victim.DUT) == 2
        assert len(self.Victim.PTB) == 2

        assert self.Victim["DUT"] is not None
        assert self.Victim["ATC"] is not None
        assert self.Victim["PTB"] is not None

    def test__INSTANCE(self):
        self.Victim.COUNT = 2
        self.Victim.CLS_PTB = Base_Device

        victim = self.Victim(1)

        assert victim.DUT == victim.DUT[1]
        assert victim.PTB == victim.PTB[1]
        try:
            victim.ATC
        except:
            pass
        else:
            assert False

        assert hasattr(victim, "DUT") is True
        assert hasattr(victim, "TB") is True

        assert len(victim.DUT) == 2
        assert len(victim.PTB) == 2

        assert victim["DUT"] is not None
        assert victim["ATC"] is not None
        assert victim["PTB"] is not None

    # -----------------------------------------------------------------------------------------------------------------
    def test__double_init(self):
        # insts is the same!
        pass

    # def reconnect


# =====================================================================================================================
